/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wProblem.h"
#include "wElement.h"
#include "wFluid.h"
#include "wBody.h"
#include "wAssign.h"
#include "wFreestream.h"
#include "wWake.h"
#include "wKutta.h"
#include "wBlowing.h"
using namespace tbox;
using namespace dart;

Problem::Problem(std::shared_ptr<MshData> _msh, int dim, double aoa, double aos,
                 double minf, double sref, double cref, double xref,
                 double yref, double zref) : msh(_msh), nDim(dim), alpha(aoa),
                                             beta(aos), M_inf(minf),
                                             S_ref(sref), c_ref(cref),
                                             x_ref(xref, yref, zref),
                                             dirD(dim, sref, aoa, aos),
                                             dirS(dim, sref, aoa, aos),
                                             dirL(dim, sref, aoa, aos)
{
    obs.push_back(&dirD);
    obs.push_back(&dirS);
    obs.push_back(&dirL);
}

/**
 * @brief Set the fluid
 */
void Problem::set(std::shared_ptr<Fluid> f)
{
    fluid = f;
    obs.push_back(f->phiInf);
}

/**
 * @brief Add a body in the fluid
 */
void Problem::add(std::shared_ptr<Body> b)
{
    bodies.push_back(b);
}

/**
 * @brief Add initial condition
 */
void Problem::set(std::shared_ptr<Initial> i)
{
    iIC = i;
    obs.push_back(i->f);
}

/**
 * @brief Add Dirichlet boundary condition
 */
void Problem::add(std::shared_ptr<Dirichlet> d)
{
    dBCs.push_back(d);
    obs.push_back(d->f);
}

/**
 * @brief Add freestream (Neumann) boundary condition
 */
void Problem::add(std::shared_ptr<Freestream> f)
{
    fBCs.push_back(f);
    obs.push_back(f->f);
}

/**
 * @brief Add wake boundary condition
 */
void Problem::add(std::shared_ptr<Wake> w)
{
    wBCs.push_back(w);
}

/**
 * @brief Add Kutta condition
 */
void Problem::add(std::shared_ptr<Kutta> k)
{
    kSCs.push_back(k);
}

/**
 * @brief Add Blowing boundary condition
 */
void Problem::add(std::shared_ptr<Blowing> b)
{
    bBCs.push_back(b);
}

/**
 * @brief Update the angle of attack
 */
void Problem::update(double aoa)
{
    alpha = aoa;
    for (auto o : obs)
        o->update(aoa);
}

/**
 * @brief Initialize the elements precomputed values
 */
void Problem::initElems()
{
    // Update volume
    for (auto e : fluid->tag->elems)
        e->initValues(true, 1);
    // Update surface (boundaries)
    for (auto surf : bodies)
        for (auto e : surf->groups[0]->tag->elems)
            e->initValues(false);
    // Update surface (Freestream B.C.)
    for (auto surf : fBCs)
        for (auto e : surf->tag->elems)
            e->initValues(false);
    // Update surface (Wake B.C.)
    for (auto surf : wBCs)
    {
        for (auto e : surf->groups[0]->tag->elems)
            e->initValues(false);
        for (auto e : surf->groups[1]->tag->elems)
            e->initValues(false);
    }
    // Update surface (Blowing B.C.) if not already done
    for (auto surf : bBCs)
        for (auto e : surf->tag->elems)
            if (!(e->hasValues()))
                e->initValues(false);
}

/**
 * @brief Initialize the elements precomputed gradients
 */
void Problem::initGradElems()
{
    // Update volume
    for (auto e : fluid->tag->elems)
        e->initGradients();
    // Update surface (boundaries)
    for (auto surf : bodies)
        for (auto e : surf->groups[0]->tag->elems)
            e->initGradients();
    // Update surface (Wake B.C.)
    for (auto surf : wBCs)
    {
        for (auto e : surf->groups[0]->tag->elems)
            e->initGradients();
        for (auto e : surf->groups[1]->tag->elems)
            e->initGradients();
    }
    // Update surface (Blowing B.C.) if not already done
    for (auto surf : bBCs)
        for (auto e : surf->tag->elems)
            if (!(e->hasGradients()))
                e->initGradients();
}

/**
 * @brief Pin a degree of freedom if no Dirichlet BC is explicitely provided
 */
void Problem::pin()
{
    if (dBCs.empty())
        this->add(std::make_shared<Dirichlet>(msh, fBCs.front()->tag->no, nDim, alpha, beta, true));
}

/**
 * @brief Check that Problem is not empty and that elements are supported
 */
void Problem::check() const
{
    // Sanity checks
    if (fluid == NULL)
        throw std::runtime_error("No Fluid provided!\n");
    if (bodies.empty())
        throw std::runtime_error("No Body provided!\n");
    if (iIC == NULL)
        throw std::runtime_error("No Initial Condition provided!\n");
    if (fBCs.empty())
        throw std::runtime_error("No Freestream B.C. provided!\n");
    if (wBCs.empty())
        throw std::runtime_error("No Wake B.C. provided!\n");
    if (kSCs.empty())
        throw std::runtime_error("No Kutta condition provided!\n");
    // Three-dimension problem
    if (nDim == 3)
    {
        // Volume element type
        for (auto e : fluid->tag->elems)
            if (e->type() != ElType::TETRA4)
                throwUnsupElType(e, "3", "volume", ElType::TETRA4);
        // Surface element type (Body)
        for (auto surf : bodies)
            for (auto e : surf->groups[0]->tag->elems)
                if (e->type() != ElType::TRI3)
                    throwUnsupElType(e, "3", "surface", ElType::TRI3);
        // Surface element type (Freestream B.C.)
        for (auto surf : fBCs)
            for (auto e : surf->tag->elems)
                if (e->type() != ElType::TRI3)
                    throwUnsupElType(e, "3", "surface", ElType::TRI3);
        // Surface element type (Wake B.C.)
        for (auto surf : wBCs)
        {
            for (auto e : surf->groups[0]->tag->elems)
                if (e->type() != ElType::TRI3)
                    throwUnsupElType(e, "3", "surface", ElType::TRI3);
            for (auto e : surf->groups[1]->tag->elems)
                if (e->type() != ElType::TRI3)
                    throwUnsupElType(e, "3", "surface", ElType::TRI3);
        }
    }
    // Two-dimension problem
    else if (nDim == 2)
    {
        // Volume element type
        for (auto e : fluid->tag->elems)
            if (e->type() != ElType::TRI3)
                throwUnsupElType(e, "2", "volume", ElType::TRI3);
        // Surface element type (Body)
        for (auto surf : bodies)
            for (auto e : surf->groups[0]->tag->elems)
                if (e->type() != ElType::LINE2)
                    throwUnsupElType(e, "2", "surface", ElType::LINE2);
        // Surface element type (Freestream B.C.)
        for (auto surf : fBCs)
            for (auto e : surf->tag->elems)
                if (e->type() != ElType::LINE2)
                    throwUnsupElType(e, "2", "surface", ElType::LINE2);
        // Surface element type (Wake B.C.)
        for (auto surf : wBCs)
        {
            for (auto e : surf->groups[0]->tag->elems)
                if (e->type() != ElType::LINE2)
                    throwUnsupElType(e, "2", "surface", ElType::LINE2);
            for (auto e : surf->groups[1]->tag->elems)
                if (e->type() != ElType::LINE2)
                    throwUnsupElType(e, "2", "surface", ElType::LINE2);
        }
    }
    else
    {
        std::stringstream err;
        err << "Solver is only implemented for 2D or 3D problems (dimension "
            << nDim << " was given)!\n";
        throw std::runtime_error(err.str());
    }
}

void Problem::throwUnsupElType(Element const *e, std::string const &pdim, std::string const &edim, ElType type) const
{
    std::stringstream err;
    err << pdim << "D solver is only implemented for "
        << edim << " elements of type " << type
        << " (" << e->type() << " was given)!\n";
    throw std::runtime_error(err.str());
}

void Problem::write(std::ostream &out) const
{
    out << "dart::Problem parameters"
        << "\n\tDimension: " << nDim
        << "\n\tAngle of attack: " << alpha
        << "\n\tAngle of sideslip: " << beta
        << "\n\tFreestream Mach number: " << M_inf
        << "\n\tReference area: " << S_ref
        << "\n\tReference length: " << c_ref
        << "\n\tReference point: [" << x_ref(0) << ", " << x_ref(1) << ", " << x_ref(2) << ']'
        << std::endl;
    out << "with"
        << "\n\t" << *fluid;
    for (auto b : bodies)
        out << "\n\t" << *b;
    out << "\n\t" << *iIC;
    for (auto d : dBCs)
        out << "\n\t" << *d;
    for (auto f : fBCs)
        out << "\n\t" << *f;
    for (auto w : wBCs)
        out << "\n\t" << *w;
    for (auto k : kSCs)
        out << "\n\t" << *k;
    for (auto b : bBCs)
        out << "\n\t" << *b;
}
