#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute the nonlinear flow around extruded 2D rae2822
# Adrien Crovato
#
# Test transonic capabilities and kutta condition
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import dart.default as floD
import fwk
from fwk.testing import *
from fwk.coloring import ccolors
import numpy as np

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 1*np.pi/180
    M_inf = 0.72
    dim = 3

    # define dimension and mesh size
    spn = 0.2 # wing span
    lgt = 6.0 # channel length
    hgt = 6.0 # channel height
    wdt = 3.0 # channel width
    c_ref = 1.0 # reference length
    S_ref = spn # reference area
    fms = 1.0 # farfield mesh size
    nms = 0.01 # nearfield mesh size

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms["msh"].start()
    pars = {'spn' : spn, 'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msN' : nms, 'msF' : fms}
    msh, gmshWriter = floD.mesh(dim, 'models/rae_25.geo', pars, ['field', 'wing', 'symmetry', 'downstream'])
    morpher = floD.morpher(msh, dim, 'wing')
    tms["msh"].stop()

    # set the problem and add fluid, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, _ = floD.problem(msh, dim, alpha, 0., M_inf, S_ref, c_ref, 0.25, 0., 0., 'wing')
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    solver = floD.newton(pbl)
    status = solver.run()
    solver.save(gmshWriter)
    tms['solver'].stop()

    # solve adjoint problem
    print(ccolors.ANSI_BLUE + 'PyGradient...' + ccolors.ANSI_RESET)
    tms['adjoint'].start()
    adjoint = floD.adjoint(solver, morpher)
    adjoint.run()
    adjoint.save(gmshWriter)
    tms['adjoint'].stop()

    # compute norm of gradient wrt to mesh
    dClX = 0
    dCdX = 0
    for n in msh.nodes:
        dClX += np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]).dot(np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]))
        dCdX += np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]).dot(np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]))
    dClX = np.sqrt(dClX)
    dCdX = np.sqrt(dCdX)
    # recover gradient wrt to AoA from mesh gradient
    drot = np.array([[-np.sin(alpha), 0, np.cos(alpha)], [0, 0, 0], [-np.cos(alpha), 0, -np.sin(alpha)]])
    dClAoA = 0
    dCdAoA = 0
    for n in bnd.nodes:
        dx = drot.dot(np.array([n.pos[0] - 0.25, n.pos[1], n.pos[2]]))
        dClAoA += np.array([adjoint.dClMsh[n.row][0], adjoint.dClMsh[n.row][1], adjoint.dClMsh[n.row][2]]).dot(dx)
        dCdAoA += np.array([adjoint.dCdMsh[n.row][0], adjoint.dCdMsh[n.row][1], adjoint.dCdMsh[n.row][2]]).dot(dx)

    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('       M    alpha       Cl       Cd       Cm')
    print('{0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f}'.format(M_inf, alpha*180/np.pi, solver.Cl, solver.Cd, solver.Cm))

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # visualize solution
    floD.initViewer(pbl)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    if status > 0:
        raise Exception(ccolors.ANSI_RED + 'Solver failed to converge!' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('iteration count', solver.nIt, 6, 1, forceabs=True))
    tests.add(CTest('CL', solver.Cl, 0.59, 5e-2))
    tests.add(CTest('CD', solver.Cd, 0.0017, 1e-2))
    tests.add(CTest('CS', solver.Cs, 0.0000, 1e-3, forceabs=True))
    tests.add(CTest('CM', solver.Cm, -0.111, 5e-2))
    tests.add(CTest('dCl/dAoA', adjoint.dClAoa, 12.9, 5e-2))
    tests.add(CTest('dCd/dAoA', adjoint.dCdAoa, 0.045, 1e-2, forceabs=True))
    tests.add(CTest('dCl_dMsh', dClX, 22.1, 5e-2))
    tests.add(CTest('dCd_dMsh', dCdX, 1.1, 1e-1))
    tests.add(CTest('dCl/dAoA (msh)', dClAoA, adjoint.dClAoa, 5e-2))
    tests.add(CTest('dCd/dAoA (msh)', dCdAoA, adjoint.dCdAoa, 1e-2, forceabs=True))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
